package jalview.datamodel.features;

import java.util.ArrayList;
import java.util.List;

/**
 * Each node of the NCList tree consists of a range, and (optionally) the NCList
 * of ranges it encloses
 *
 * @param <V>
 */
class NCNode<V extends ContiguousI>
{
  /*
   * deep size (number of ranges included)
   */
  private int size;

  private V region;

  /*
   * null, or an object holding contained subregions of this nodes region
   */
  private NCList<V> subregions;

  /**
   * Constructor given a list of ranges
   * 
   * @param ranges
   */
  NCNode(List<V> ranges)
  {
    build(ranges);
  }

  /**
   * Constructor given a single range
   * 
   * @param range
   */
  NCNode(V range)
  {
    List<V> ranges = new ArrayList<V>();
    ranges.add(range);
    build(ranges);
  }

  NCNode(V entry, NCList<V> newNCList)
  {
    region = entry;
    subregions = newNCList;
    size = 1 + newNCList.getSize();
  }

  /**
   * @param ranges
   */
  protected void build(List<V> ranges)
  {
    size = ranges.size();

    if (!ranges.isEmpty())
    {
      region = ranges.get(0);
    }
    if (ranges.size() > 1)
    {
      subregions = new NCList<V>(ranges.subList(1, ranges.size()));
    }
  }

  int getStart()
  {
    return region.getBegin();
  }

  int getEnd()
  {
    return region.getEnd();
  }

  /**
   * Formats the node as a bracketed list e.g.
   * 
   * <pre>
   * [1-100 [10-30 [10-20]], 15-30 [20-20]]
   * </pre>
   */
  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder(10 * size);
    sb.append(region.getBegin()).append("-").append(region.getEnd());
    if (subregions != null)
    {
      sb.append(" ").append(subregions.toString());
    }
    return sb.toString();
  }

  void prettyPrint(StringBuilder sb, int offset, int indent) {
    for (int i = 0 ; i < offset ; i++) {
      sb.append(" ");
    }
    sb.append(region.getBegin()).append("-").append(region.getEnd());
    if (subregions != null)
    {
      sb.append(System.lineSeparator());
      subregions.prettyPrint(sb, offset + 2, indent);
    }
  }
  /**
   * Add any ranges that overlap the from-to range to the result list
   * 
   * @param from
   * @param to
   * @param result
   */
  void findOverlaps(long from, long to, List<V> result)
  {
    if (region.getBegin() <= to && region.getEnd() >= from)
    {
      result.add(region);
    }
    if (subregions != null)
    {
      subregions.findOverlaps(from, to, result);
    }
  }

  /**
   * Add one range to this subrange
   * 
   * @param entry
   */
  synchronized void add(V entry)
  {
    if (entry.getBegin() < region.getBegin() || entry.getEnd() > region.getEnd()) {
      throw new IllegalArgumentException(String.format(
              "adding improper subrange %d-%d to range %d-%d",
              entry.getBegin(), entry.getEnd(), region.getBegin(),
              region.getEnd()));
    }
    if (subregions == null)
    {
      subregions = new NCList<V>(entry);
    }
    else
    {
      subregions.add(entry);
    }
  }

  /**
   * Answers true if the data held satisfy the rules of construction of an
   * NCList, else false.
   * 
   * @return
   */
  boolean isValid()
  {
    if (subregions == null)
    {
      return true;
    }
    return subregions.isValid(getStart(), getEnd());
  }

  /**
   * Adds all contained entries to the given list
   * 
   * @param entries
   */
  void getEntries(List<V> entries)
  {
    entries.add(region);
    if (subregions != null)
    {
      subregions.getEntries(entries);
    }
  }
}
